package get

import (
	"fmt"
	"slices"
	"strings"

	"github.com/goccy/go-yaml"
	"github.com/spf13/cobra"

	"gitea.libretechconsulting.com/50W/eia-api-go/cmd/eia-client/internal/util"
	"gitea.libretechconsulting.com/50W/eia-api-go/pkg/eia"
)

var (
	filteredSuffixes = []string{
		"Facet",
		"FacetId",
		"Body",
	}
	filteredPrefixes = []string{
		"EIAA",
	}
)

var GetRouteCmd = &cobra.Command{
	Use:               "route",
	Aliases:           []string{"r"},
	Short:             "Describe a route",
	Args:              cobra.ExactArgs(1),
	ValidArgsFunction: completeRouteRoutes,
	Run:               RunGetRouteCmd,
}

func RunGetRouteCmd(cmd *cobra.Command, args []string) {
	logger := util.Logger(cmd)
	route := args[0]
	logger.Info().Str("route", route).Msg("getting route description")

	routeType, err := util.GetRouteType(cmd, route)
	if err != nil {
		logger.Fatal().Err(err).Send()
	}

	switch routeType {
	case eia.RouteTypeFinal:
		showFinalRoute(cmd, route)
	case eia.RouteTypeRoutes:
		showRoutes(cmd, route)
	default:
		logger.Fatal().Any("type", routeType).
			Msg("not a valid route type for get route command")
	}
}

func showFinalRoute(cmd *cobra.Command, route string) {
	finalRoute, _, err := util.GetRoute(cmd, route)
	if err != nil {
		util.Logger(cmd).Fatal().Err(err).Send()
	}

	bytes, _ := yaml.Marshal(finalRoute)
	showRoute(bytes, route, eia.RouteTypeFinal)
}

func showRoutes(cmd *cobra.Command, route string) {
	_, routes, err := util.GetRoute(cmd, route)
	if err != nil {
		util.Logger(cmd).Fatal().Err(err).Send()
	}

	bytes, _ := yaml.Marshal(routes)
	showRoute(bytes, route, eia.RouteTypeRoutes)
}

func showRoute(bytes []byte, name string, routeType eia.RouteType) {
	fmt.Println(string(util.PrettyBytes(&util.PrettyOpts{
		Bytes: append([]byte(fmt.Sprintf("name: %s\ntype: %s\n", name, routeType.String())), bytes...),
	})))
}

func completeRouteRoutes(cmd *cobra.Command, args []string, toComplete string) (
	[]string, cobra.ShellCompDirective,
) {
	routes := eia.GetRoutes("Data")
	routeRoutes := make([]string, 0, len(routes))
	for _, r := range routes {
		var rName string
		for _, suffix := range filteredSuffixes {
			if strings.HasSuffix(r, suffix) {
				goto next
			}
		}
		for _, prefix := range filteredPrefixes {
			if strings.HasPrefix(r, prefix) {
				goto next
			}
		}
		rName, _ = strings.CutSuffix(r, "Data")
		routeRoutes = append(routeRoutes, rName)
	next:
	}

	filteredRoutes := make([]string, 0, len(routeRoutes))
	if toComplete != "" {
		for _, route := range routeRoutes {
			if strings.HasPrefix(route, toComplete) {
				filteredRoutes = append(filteredRoutes, route)
			}
		}
	} else {
		filteredRoutes = routeRoutes
	}
	return slices.Clip(filteredRoutes), cobra.ShellCompDirectiveNoFileComp
}
